import json
import random
from typing import Tuple, Dict, Any, Optional

from fastchat.model import load_model, get_conversation_template, add_model_args
from ftfy import fix_text
import torch
from transformers.generation import GenerationConfig

from rank_llm.rerank.rankllm import RankLLM, PromptMode
from rank_llm.result import Result


class RankListwiseOSLLM(RankLLM):
    def __init__(
        self,
        model: str,
        context_size: int = 4096,
        prompt_mode: PromptMode = PromptMode.RANK_GPT,
        num_few_shot_examples: int = 0,
        device: str = "cuda",
        num_gpus: int = 1,
        variable_passages: bool = False,
        window_size: int = 20,
        system_message: str = None,
    ) -> None:
        super().__init__(model, context_size, prompt_mode, num_few_shot_examples)
        self._device = device
        if self._device == "cuda":
            assert torch.cuda.is_available()
        if prompt_mode != PromptMode.RANK_GPT:
            raise ValueError(
                f"Unsupported prompt mode: {prompt_mode}. The only prompt mode currently supported is a slight variation of Rank_GPT prompt."
            )
        # ToDo: Make repetition_penalty configurable
        self._llm, self._tokenizer = load_model(model, device=device, num_gpus=num_gpus)
        self._variable_passages = variable_passages
        self._window_size = window_size
        self._system_message = system_message
        self._output_token_estimate = None
        if num_few_shot_examples > 0:
            with open("data/output_v2_aug_filtered.jsonl", "r") as json_file:
                self._examples = list(json_file)[1:-1]

    def run_llm(
        self, prompt: str, current_window_size: Optional[int] = None
    ) -> Tuple[str, int]:
        if current_window_size is None:
            current_window_size = self._window_size
        inputs = self._tokenizer([prompt])
        inputs = {k: torch.tensor(v).to(self._device) for k, v in inputs.items()}
        gen_cfg = GenerationConfig.from_model_config(self._llm.config)
        gen_cfg.max_new_tokens = self.num_output_tokens(current_window_size)
        gen_cfg.min_new_tokens = self.num_output_tokens(current_window_size)
        # gen_cfg.temperature = 0
        gen_cfg.do_sample = False
        output_ids = self._llm.generate(**inputs, generation_config=gen_cfg)

        if self._llm.config.is_encoder_decoder:
            output_ids = output_ids[0]
        else:
            output_ids = output_ids[0][len(inputs["input_ids"][0]) :]
        outputs = self._tokenizer.decode(
            output_ids, skip_special_tokens=True, spaces_between_special_tokens=False
        )
        return outputs, output_ids.size(0)

    def num_output_tokens(self, current_window_size: Optional[int] = None) -> int:
        if current_window_size is None:
            current_window_size = self._window_size
        if self._output_token_estimate and self._window_size == current_window_size:
            return self._output_token_estimate
        else:
            _output_token_estimate = (
                len(
                    self._tokenizer.encode(
                        " > ".join([f"[{i+1}]" for i in range(current_window_size)])
                    )
                )
                - 1
            )
            if (
                self._output_token_estimate is None
                and self._window_size == current_window_size
            ):
                self._output_token_estimate = _output_token_estimate
            return _output_token_estimate

    def _add_prefix_prompt(self, query: str, num: int) -> str:
        return f"I will provide you with {num} passages, each indicated by a numerical identifier []. Rank the passages based on their relevance to the search query: {query}.\n"

    def _add_post_prompt(self, query: str, num: int) -> str:
        example_ordering = "[2] > [1]" if self._variable_passages else "[4] > [2]"
        return f"Search Query: {query}.\nRank the {num} passages above based on their relevance to the search query. All the passages should be included and listed using identifiers, in descending order of relevance. The output format should be [] > [], e.g., {example_ordering}, Only respond with the ranking results, do not say any word or explain."

    def _add_few_shot_examples(self, conv):
        for _ in range(self._num_few_shot_examples):
            ex = random.choice(self._examples)
            obj = json.loads(ex)
            prompt = obj["conversations"][0]["value"]
            response = obj["conversations"][1]["value"]
            conv.append_message(conv.roles[0], prompt)
            conv.append_message(conv.roles[1], response)
        return conv

    def create_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[str, int]:
        query = result.query
        num = len(result.hits[rank_start:rank_end])
        max_length = 300 * (20 / (rank_end - rank_start))
        while True:
            conv = get_conversation_template(self._model)
            if self._system_message:
                conv.set_system_message(self._system_message)
            conv = self._add_few_shot_examples(conv)
            prefix = self._add_prefix_prompt(query, num)
            rank = 0
            input_context = f"{prefix}\n"
            for hit in result.hits[rank_start:rank_end]:
                rank += 1
                content = hit["content"]
                content = content.replace("Title: Content: ", "")
                content = content.strip()
                # For Japanese should cut by character: content = content[:int(max_length)]
                content = " ".join(content.split()[: int(max_length)])
                input_context += f"[{rank}] {self._replace_number(content)}\n"

            input_context += self._add_post_prompt(query, num)
            conv.append_message(conv.roles[0], input_context)
            conv.append_message(conv.roles[1], None)
            prompt = conv.get_prompt()
            prompt = fix_text(prompt)
            num_tokens = self.get_num_tokens(prompt)
            if num_tokens <= self.max_tokens() - self.num_output_tokens(
                rank_end - rank_start
            ):
                break
            else:
                max_length -= max(
                    1,
                    (
                        num_tokens
                        - self.max_tokens()
                        + self.num_output_tokens(rank_end - rank_start)
                    )
                    // ((rank_end - rank_start) * 4),
                )
        return prompt, self.get_num_tokens(prompt)

    def get_num_tokens(self, prompt: str) -> int:
        return len(self._tokenizer.encode(prompt))

    def cost_per_1k_token(self, input_token: bool) -> float:
        return 0



INDICES = {
    "bm25": {
        "dl19": "msmarco-v1-passage",
        "dl20": "msmarco-v1-passage",
        "dl21": "msmarco-v2-passage",
        "dl22": "msmarco-v2-passage",
        "covid": "beir-v1.0.0-trec-covid.flat",
        "arguana": "beir-v1.0.0-arguana.flat",
        "touche": "beir-v1.0.0-webis-touche2020.flat",
        "news": "beir-v1.0.0-trec-news.flat",
        "scifact": "beir-v1.0.0-scifact.flat",
        "fiqa": "beir-v1.0.0-fiqa.flat",
        "scidocs": "beir-v1.0.0-scidocs.flat",
        "nfc": "beir-v1.0.0-nfcorpus.flat",
        "quora": "beir-v1.0.0-quora.flat",
        "dbpedia": "beir-v1.0.0-dbpedia-entity.flat",
        "fever": "beir-v1.0.0-fever.flat",
        "robust04": "beir-v1.0.0-robust04.flat",
        "signal": "beir-v1.0.0-signal1m.flat",
        "mrtydi-ar": "mrtydi-v1.1-arabic",
        "mrtydi-bn": "mrtydi-v1.1-bengali",
        "mrtydi-fi": "mrtydi-v1.1-finnish",
        "mrtydi-id": "mrtydi-v1.1-indonesian",
        "mrtydi-ja": "mrtydi-v1.1-japanese",
        "mrtydi-ko": "mrtydi-v1.1-korean",
        "mrtydi-ru": "mrtydi-v1.1-russian",
        "mrtydi-sw": "mrtydi-v1.1-swahili",
        "mrtydi-te": "mrtydi-v1.1-telugu",
        "mrtydi-th": "mrtydi-v1.1-thai",
    },
    "SPLADE++_EnsembleDistil_ONNX": {
        "dl19": "msmarco-v1-passage-splade-pp-ed-text",
        "dl20": "msmarco-v1-passage-splade-pp-ed-text",
        "dl21": "",
        "dl22": "",
        "covid": "",
        "arguana": "",
        "touche": "",
        "news": "",
        "scifact": "",
        "fiqa": "",
        "scidocs": "",
        "nfc": "",
        "quora": "",
        "dbpedia": "",
        "fever": "",
        "robust04": "",
        "signal": "",
    },
    "distilbert_tas_b": {
        "dl19": "msmarco-v1-passage.distilbert-dot-tas_b-b256",
        "dl20": "msmarco-v1-passage.distilbert-dot-tas_b-b256",
        "dl21": "",
        "dl22": "",
        "covid": "",
        "arguana": "",
        "touche": "",
        "news": "",
        "scifact": "",
        "fiqa": "",
        "scidocs": "",
        "nfc": "",
        "quora": "",
        "dbpedia": "",
        "fever": "",
        "robust04": "",
        "signal": "",
    },
    "openai-ada2": {
        "dl19": "msmarco-v1-passage.openai-ada2",
        "dl20": "msmarco-v1-passage.openai-ada2",
        "dl21": "",
        "dl22": "",
        "covid": "",
        "arguana": "",
        "touche": "",
        "news": "",
        "scifact": "",
        "fiqa": "",
        "scidocs": "",
        "nfc": "",
        "quora": "",
        "dbpedia": "",
        "fever": "",
        "robust04": "",
        "signal": "",
    },
}



from enum import Enum
import json
from pathlib import Path
from typing import List, Union, Dict, Any

from rank_llm.retrieve.pyserini_retriever import PyseriniRetriever
from rank_llm.retrieve.pyserini_retriever import RetrievalMethod
from rank_llm.result import Result


class RetrievalMode(Enum):
    DATASET = "dataset"
    QUERY_AND_DOCUMENTS = "query_and_documents"
    QUERY_AND_HITS = "query_and_hits"
    SAVED_FILE = "saved_file"

    def __str__(self):
        return self.value


class Retriever:
    def __init__(
        self,
        retrieval_mode: RetrievalMode,
        dataset: Union[str, List[str], List[Dict[str, Any]]],
        retrieval_method: RetrievalMethod = RetrievalMethod.UNSPECIFIED,
        query: str = None,
    ) -> None:
        self._retrieval_mode = retrieval_mode
        self._dataset = dataset
        self._retrieval_method = retrieval_method
        self._query = query

    @staticmethod
    def from_inline_documents(query: str, documents: List[str]):
        """
        Creates a Retriever instance for inline documents with the passed in query.

        Args:
            query (str): The search query.
            documents (List[str]): A list of documents to search from.

        Returns:
            List[Dict[str, Any]]: The retrieval results.

        Raises:
            ValueError: If query or documents are invalid or missing.
        """
        if not query or query == "":
            raise ValueError("Please provide a query string.")
        if not documents:
            raise ValueError("Please provide a non-empty list of documents.")
        retriever = Retriever(
            RetrievalMode.QUERY_AND_DOCUMENTS, dataset=documents, query=query
        )
        return retriever.retrieve()

    @staticmethod
    def from_inline_hits(query: str, hits: List[Dict[str, Any]]):
        """
        Creates a Retriever instance for inline hits with the passed in query.

        Args:
            query (str): The search query.
            hits (List[Dict[str, Any]]): Predefined hits relevant to the query.

        Returns:
            List[Dict[str, Any]]: The retrieval results.

        Raises:
            ValueError: If query or hits are invalid or missing.
        """
        if not query or query == "":
            raise ValueError("Please provide a query string.")
        if not hits:
            raise ValueError("Please provide a non-empty list of hits.")

        retriever = Retriever(RetrievalMode.QUERY_AND_HITS, dataset=hits, query=query)
        return retriever.retrieve()

    @staticmethod
    def from_dataset_with_prebuit_index(
        dataset_name: str, retrieval_method: RetrievalMethod = RetrievalMethod.BM25
    ):
        """
        Creates a Retriever instance for a dataset with a prebuilt index.

        Args:
            dataset_name (str): The name of the dataset.
            retrieval_method (RetrievalMethod): The retrieval method to be used (e.g. BM25).

        Returns:
            List[Dict[str, Any]]: The retrieval results.

        Raises:
            ValueError: If dataset name or retrieval method is invalid or missing.
        """
        if not dataset_name:
            raise ValueError("Please provide name of the dataset.")
        if not isinstance(dataset_name, str):
            raise ValueError(
                f"Invalid dataset format: {dataset_name}. Expected a string representing name of the dataset."
            )
        if not retrieval_method:
            raise "Please provide a retrieval method."
        if retrieval_method == RetrievalMethod.UNSPECIFIED:
            raise ValueError(
                f"Invalid retrieval method: {retrieval_method}. Please provide a specific retrieval method."
            )
        retriever = Retriever(
            RetrievalMode.DATASET,
            dataset=dataset_name,
            retrieval_method=retrieval_method,
        )
        return retriever.retrieve()

    @staticmethod
    def from_saved_results(file_name: str):
        """
        Creates a Retriever instance from saved retrieval results specified by 'file_name'.

        Args:
            file_name (str): The file name containing the saved retrieval results.

        Returns:
            List[Dict[str, Any]]: The retrieval results loaded from the file.

        Raises:
            ValueError: If the file content is not in the expected format.
        """
        with open(file_name, "r") as f:
            retrieved_results = json.load(f)
        if not isinstance(retrieved_results, list):
            raise ValueError(
                f"Invalid retrieval format: Expected a list of dictionaries, got {type(retrieved_results)}"
            )
        retriever = Retriever(RetrievalMode.SAVED_FILE, dataset=retrieved_results)
        return retriever.retrieve()

    def retrieve(self) -> List[Dict[str, Any]]:
        """
        Executes the retrieval process based on the configation provided with the Retriever instance.

        Returns:
            List[Dict[str, Any]]: A list of retrieval results.

        Raises:
            ValueError: If the retrieval mode is invalid or the result format is not as expected.
        """
        if self._retrieval_mode == RetrievalMode.DATASET:
            candidates_file = Path(
                f"retrieve_results/{self._retrieval_method.name}/retrieve_results_{self._dataset}.json"
            )
            if not candidates_file.is_file():
                print(f"Retrieving with dataset {self._dataset}")
                pyserini = PyseriniRetriever(self._dataset, self._retrieval_method)
                # Always retrieve top 100 so that results are reusable for all top_k_candidates values.
                retrieved_results = pyserini.retrieve_and_store(k=100)
            else:
                print("Reusing existing retrieved results.")
                with open(candidates_file, "r") as f:
                    loaded_results = json.load(f)
                retrieved_results = [
                    Result(r["query"], r["hits"]) for r in loaded_results
                ]

        elif self._retrieval_mode == RetrievalMode.QUERY_AND_DOCUMENTS:
            document_hits = []
            for i, document in enumerate(self._dataset):
                if not isinstance(document, str):
                    raise ValueError(
                        f"Invalid dataset format: {self._dataset}. Expected a list of strings where each string represents a document."
                    )
                document_hits.append(
                    {"content": document, "qid": 1, "docid": i + 1, "rank": i + 1}
                )
            retrieved_results = [Result(self._query, document_hits)]

        elif self._retrieval_mode == RetrievalMode.QUERY_AND_HITS:
            retrieved_results = [Result(self._query, self._dataset)]
        elif self._retrieval_mode == RetrievalMode.SAVED_FILE:
            retrieved_results = [
                Result(query=r["query"], hits=r["hits"]) for r in self._dataset
            ]
        else:
            raise ValueError(f"Invalid retrieval mode: {self._retrieval_mode}")
        for result in retrieved_results:
            self._validate_result(result)
        return retrieved_results

    def _validate_result(self, result: Result):
        if not isinstance(result, Result):
            raise ValueError(
                f"Invalid result format: Expected type `Result`, got {type(result)}"
            )
        if not result.query:
            raise ValueError(f"Invalid format: missing `query`")
        if not result.hits:
            raise ValueError(f"Invalid format: missing `hits`")
        for hit in result.hits:
            if not isinstance(hit, Dict):
                raise ValueError(
                    f"Invalid hits format: Expected a list of Dicts where each Dict represents a hit."
                )
            if "content" not in hit.keys():
                raise ValueError((f"Invalid format: Missing `content` key in hit"))
